# -*- coding: utf-8 -*-
from maryjane.tags.base import DictionaryTag

__author__ = 'vahid'


class ContextTag(DictionaryTag):
    def __init__(self, manifest, **attributes):
        super(ContextTag, self).__init__(manifest, **attributes)
        manifest.context.update(attributes)
        manifest.context.update(self.to_dict())



# -*- coding: utf-8 -*-
from maryjane.tags.base import DictionaryTag
from maryjane.helpers import split_paths
from watchdog.events import FileSystemEventHandler
import traceback
__author__ = 'vahid'


class TaskTag(DictionaryTag):
    priority = 50

    def __init__(self, manifest, **attributes):
        if 'priority' in attributes:
            attributes['priority'] = int(attributes['priority'])
        super(TaskTag, self).__init__(manifest, **attributes)


    def execute_actions(self):
        if hasattr(self, 'banner'):
            print self.banner
        for action in self.actions:
            action.execute()

    def __repr__(self):
        return '<TaskTag>'

class TaskEventHandler(FileSystemEventHandler):
    def __init__(self, task):
        self.task = task
        super(FileSystemEventHandler, self).__init__()

    def on_any_event(self, event):
        # noinspection PyBroadException
        try:
            self.task.execute_if_needed(event)
        except:
            traceback.print_exc()

class ObservableTaskTag(TaskTag):

    def execute_if_needed(self, event):
        update_needed = False
        paths = []
        if hasattr(event, 'src_path'):
            paths += split_paths(event.src_path)
        if hasattr(event, 'dest_path'):
            paths += split_paths(event.dest_path)

        if self.watcher.predicate:
            for p in paths:
                if self.watcher.predicate(p):
                    update_needed = True
                    break
        else:
            update_needed = True


        if update_needed:
            self.execute_actions()

    def create_event_handler(self):
        return TaskEventHandler(self)

    def __repr__(self):
        return '<ObservableTaskTag>'



# -*- coding: utf-8 -*-

from maryjane.tags.action import ActionTag
from maryjane.tags.options import OptionsTag
import subprocess as sb

__author__ = 'vahid'
class SubprocessActionTag(ActionTag):

    def __init__(self, manifest, popen_args, **attributes):
        self.popen_args = popen_args
        super(SubprocessActionTag, self).__init__(manifest, **attributes)

    @classmethod
    def from_yaml_node(cls, manifest, loader, node):
        kw = loader.construct_mapping(node)

        class_kwargs = {}
        for arg_name in ('arguments', 'banner'):
            arg_value = kw.get(arg_name)
            if arg_value:
                class_kwargs[arg_name] = arg_value
                del kw[arg_name]

        popen_args = OptionsTag(manifest, **kw)

        return cls(manifest, popen_args, **class_kwargs)

    def execute(self):
        super(SubprocessActionTag, self).execute()
        args = self.popen_args.to_dict()
        if 'manifest' in args:
            del args['manifest']
        p = sb.Popen(self.arguments, **args)
        p.wait()

    def __repr__(self):
        return '<SubprocessActionTag %s>' % self.executable




# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod

__author__ = 'vahid'

class BaseTag(object):
    __metaclass__ = ABCMeta

    def __init__(self, manifest, **kw):
        self.manifest = manifest

    def __getattribute__(self, item):
        v = object.__getattribute__(self, item)
        if isinstance(v, LazyTag):
            return v.lazy_value()
        elif isinstance(v, list):
            return [i if not isinstance(i, LazyTag) else i.lazy_value() for i in v]
        return v

class DictionaryTag(BaseTag):

    def __init__(self, manifest, **attributes):
        super(DictionaryTag, self).__init__(manifest)
        self.__dict__.update(attributes)

    def to_dict(self):
        return {k: getattr(self, k) for k in self.__dict__ if not k.startswith('_')}

    @classmethod
    def from_yaml_node(cls, manifest, loader, node):
        kw = loader.construct_mapping(node)
        return cls(manifest, **kw)

    def __repr__(self):
        return '<DictionaryTag>'


class ScalarTag(BaseTag):

    def __init__(self, value, manifest):
        self._value = value
        BaseTag.__init__(self, manifest)

    @property
    def value(self):
        return self._value

    @classmethod
    def from_yaml_node(cls, manifest, loader, node):
        value = loader.construct_scalar(node)
        return cls(value, manifest)


class LazyTag(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def lazy_value(self):
        pass


class LazyScalarTag(ScalarTag, LazyTag):
    __metaclass__ = ABCMeta

    @abstractmethod
    def lazy_value(self):
        pass




# -*- coding: utf-8 -*-

from .base import LazyTag, BaseTag, ScalarTag, LazyScalarTag, DictionaryTag
from .options import OptionsTag
from .task import TaskTag, ObservableTaskTag
from .template import TemplateTag
from .subprocessing import SubprocessActionTag
from .evaluate import EvaluateTag
from .watcher import WatcherTag
from .pyimport import ImportTag
from .banner import BannerActionTag
from .context import ContextTag

__author__ = 'vahid'



# -*- coding: utf-8 -*-
from maryjane.tags.base import DictionaryTag
__author__ = 'vahid'


class WatcherTag(DictionaryTag):
    recursive = False
    strict = True
    predicate = None

    def __repr__(self):
        return '<WatcherTag>'




# -*- coding: utf-8 -*-
from maryjane.tags.base import DictionaryTag


__author__ = 'vahid'

class ActionTag(DictionaryTag):

    def execute(self):
        if hasattr(self, 'banner'):
            print self.banner

    def __repr__(self):
        return '<ActionTag>'




# -*- coding: utf-8 -*-
import yaml
import time
import os.path
import traceback
from collections import OrderedDict
from watchdog.observers import Observer
from maryjane.tags import TaskTag, SubprocessActionTag, TemplateTag, ObservableTaskTag,\
    EvaluateTag, OptionsTag, WatcherTag, ImportTag, BannerActionTag, ContextTag
from maryjane.helpers import get_source_dirs, has_file_overlap, split_paths, get_filename
from watchdog.events import FileSystemEventHandler
__author__ = 'vahid'


class ManifestFileEventHandler(FileSystemEventHandler):
    def __init__(self, manifest):
        self.manifest = manifest
        super(FileSystemEventHandler, self).__init__()

    def on_any_event(self, event):
        paths = []
        if hasattr(event, 'src_path'):
            paths += split_paths(event.src_path)
        if hasattr(event, 'dest_path'):
            paths += split_paths(event.dest_path)

        if has_file_overlap(paths, self.manifest.filename):
            # noinspection PyBroadException
            try:
                self.manifest.reboot()
            except:
                traceback.print_exc()

_context_builtins = {
    'has_file_overlap': has_file_overlap,
    'split_paths': split_paths,
    'get_source_dirs': get_source_dirs,
    'get_filename': get_filename
}

class ManifestObserver(Observer):

    def __init__(self, manifest):
        self.manifest = manifest
        Observer.__init__(self)
        self.schedule(ManifestFileEventHandler(self.manifest),
                      get_source_dirs(self.manifest.filename)[0])

    def run(self):
        print "Starting Manifest Watcher"
        self.manifest.watch()
        super(ManifestObserver, self).run()

class Manifest(object):

    def __init__(self, filename, working_dir='.'):
        self.filename = filename
        self.working_dir = working_dir
        self.configure_yaml()
        self.tasks = {}
        self.watching_tasks = {}
        self._context = {}
        self.observer = None
        self.is_watching = False
        self.reload_file()

    def reboot(self):
        self.unwatch()
        self.reload_file()
        self.execute()
        self.watch()

    def reload_file(self):
        manifest_dir = os.path.dirname(self.filename)
        self._context = dict(manifest_dir= '.' if not manifest_dir else manifest_dir,
                             working_dir=self.working_dir)
        self._context.update(_context_builtins)
        with open(self.filename) as manifest_file:
            config = yaml.load(manifest_file)

        self.tasks = OrderedDict(sorted([(k, v) for k, v in config.iteritems() if isinstance(v, TaskTag)],
                                        key=lambda i: i[1].priority))

        self.watching_tasks = {k: v for k, v in config.iteritems() if isinstance(v, ObservableTaskTag)}
        # for context_key, context in {k: v for k, v in config.iteritems() if isinstance(v, ContextTag)}.iteritems():
        #     self._context.update(context.to_dict())
        self._context.update(self.tasks)


    def __getattr__(self, name):
        if name in self.tasks:
            return self.tasks[name]
        raise AttributeError

    @property
    def context(self):
        return self._context

    def execute(self):
        for task_name, task in self.tasks.iteritems():
            task.execute_actions()

    def configure_yaml(self):
        def specialize(func):
            def _decorator(*args, **kwargs):
                return func(self, *args, **kwargs)
            return _decorator
        yaml.add_constructor('!options', specialize(OptionsTag.from_yaml_node))
        yaml.add_constructor('!subprocess', specialize(SubprocessActionTag.from_yaml_node))
        yaml.add_constructor('!task', specialize(TaskTag.from_yaml_node))
        yaml.add_constructor('!watch', specialize(ObservableTaskTag.from_yaml_node))
        yaml.add_constructor('!template', specialize(TemplateTag.from_yaml_node))
        yaml.add_constructor('!eval', specialize(EvaluateTag.from_yaml_node))
        yaml.add_constructor('!watcher', specialize(WatcherTag.from_yaml_node))
        yaml.add_constructor('!import', specialize(ImportTag.from_yaml_node))
        yaml.add_constructor('!banner', specialize(BannerActionTag.from_yaml_node))
        yaml.add_constructor('!context', specialize(ContextTag.from_yaml_node))



    def unwatch(self):
        if self.observer.is_alive():
            self.observer.stop()

    def watch(self):
        self.observer = Observer()
        for task_name, task in self.watching_tasks.iteritems():
            handler = task.create_event_handler()
            for directory in get_source_dirs(task.watcher.sources):
                self.observer.schedule(handler, directory, recursive=task.watcher.recursive)
        print "Starting Task Watcher"
        self.observer.start()


if __name__ == '__main__':
    fn = 'tests/maryjane.yaml'
    m = Manifest(fn, working_dir='tests')
    m.execute()
    #m.watch()
    o = ManifestObserver(m)
    o.start()
    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            o.stop()
            o.join()
            break



# -*- coding: utf-8 -*-
from maryjane.tags import LazyScalarTag
from mako.template import Template as MakoTemplate
__author__ = 'vahid'


class TemplateTag(LazyScalarTag, MakoTemplate):

    def __init__(self, template, manifest):
        LazyScalarTag.__init__(self, template, manifest)
        MakoTemplate.__init__(self, self.value, strict_undefined=True)

    def lazy_value(self):
        return self.render(**self.manifest.context)


    def __repr__(self):
        return '<TemplateTag "%s">' % self.source




# -*- coding: utf-8 -*-

from maryjane.tags.action import ActionTag

__author__ = 'vahid'

class BannerActionTag(ActionTag):

    def __init__(self, manifest, **attributes):
        super(BannerActionTag, self).__init__(manifest, **attributes)

    def execute(self):
        super(BannerActionTag, self).execute()
        print self.text

    def __repr__(self):
        return '<BannerActionTag %s>' % self.executable




# -*- coding: utf-8 -*-
from maryjane.tags import LazyScalarTag
__author__ = 'vahid'


class EvaluateTag(LazyScalarTag):
    def __init__(self, raw_value, manifest):
        super(EvaluateTag, self).__init__(raw_value ,manifest)

    def lazy_value(self):
        return eval(self.value, self.manifest.context)




# -*- coding: utf-8 -*-
import os
import glob

__author__ = 'vahid'


def split_paths(p):
    if isinstance(p, basestring):
        # noinspection PyUnresolvedReferences
        res = p.split(',')
    else:
        res = p
    return [os.path.abspath(i.strip()) for i in res]


def distinct(l):
    if not l or not len(l):
        return []
    return list(set(l))


def get_source_files(sources, extensions='.sass,.scss'):
    source_files = []
    for s in sources:
        p = os.path.abspath(s)
        if os.path.isfile(p):
            source_files.append(p)
        elif os.path.isdir(p):
            for extension in extensions.split(','):
                for _f in glob.iglob(os.path.join(p,'*%s' % extension)):
                    source_files.append(os.path.abspath(_f))
    return sorted(distinct(source_files))


def get_source_dirs(sources):
    source_dirs = []
    for s in split_paths(sources):
        if os.path.isdir(s):
            source_dirs.append(s)
        else:
            p = os.path.abspath(os.path.dirname(s))
            if os.path.isdir(p):
                source_dirs.append(p)

    return sorted(distinct(source_dirs))


def has_file_overlap(paths1, paths2):
    for p1 in split_paths(paths1):
        if not os.path.isfile(p1):
            continue
        for p2 in split_paths(paths2):
            if not os.path.isfile(p2):
                continue
            if os.path.samefile(p1, p2):
                return True
    return False

def get_filename(filepath):
    return os.path.split(filepath)[1]



# -*- coding: utf-8 -*-

from .manifest import Manifest, ManifestObserver
import time
import os
import traceback
__author__ = 'vahid'
__version__ = '2.16'


def watch(manifest_to_watch, block=False):
    observer = ManifestObserver(manifest_to_watch)
    observer.start()
    if block:
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                observer.stop()
                observer.join()
                break


def main(manifest_filename, enable_watcher=False, block=False, working_directory='.'):
    m = Manifest(manifest_filename, working_dir=os.path.abspath(working_directory))
    try:
        m.execute()
    except:
        traceback.print_stack()
        traceback.print_exc()
    if enable_watcher:
        watch(m, block=block)
    return m




# -*- coding: utf-8 -*-
"""
Created on:    Nov 1, 2013
@author:        vahid
"""

from setuptools import setup, find_packages
import os.path
import re

# reading isass version (same way sqlalchemy does)
with open(os.path.join(os.path.dirname(__file__), 'maryjane', '__init__.py')) as v_file:
    package_version = re.compile(r".*__version__ = '(.*?)'", re.S).match(v_file.read()).group(1)

long_description = """
`MaryJane` is a Python automatic file merge tool.
Docs at http://github.com/pylover/maryjane.
"""

setup(
    name='maryjane',
    version=package_version,
    author='Vahid Mardani',
    author_email='vahid.mardani@gmail.com',
    url='http://github.com/pylover/MaryJane',
    description='Python file watcher and task manager',
    long_description=long_description,
    license='MIT',
    install_requires=['mako',
                      'pyyaml',
                      'watchdog>=0.6.0'],
    packages=['maryjane', 'maryjane.tags'],
    entry_points={
        'console_scripts': [
            'maryjane = maryjane.cli:cli_main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Communications :: Email',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        ],
    )


# -*- coding: utf-8 -*-
__author__ = 'vahid'
import argparse
import os
import sys
from maryjane import main

parser = argparse.ArgumentParser(description='File watcher and task manager')
parser.add_argument('manifest', nargs='?', default='maryjane.yaml', help='Manifest file, default: "maryjane.yaml"')
parser.add_argument('-d', '--working-directory', metavar='WORKING_DIR', default='.', dest='working_dir', help='Working directory, default: current directory: %s' % os.path.abspath(os.curdir))
parser.add_argument('-w','--watch', dest='watch', action='store_true', help='Watch for modifications, and execute tasks if needed.')

def cli_main():
    args = parser.parse_args()
    main(args.manifest,
         enable_watcher=args.watch,
         block=args.watch,
         working_directory=args.working_dir)
    sys.exit(0)



# -*- coding: utf-8 -*-
from maryjane.tags import LazyScalarTag
import importlib
__author__ = 'vahid'


class ImportTag(LazyScalarTag):
    def __init__(self, raw_value, manifest):
        super(ImportTag, self).__init__(raw_value ,manifest)

    def lazy_value(self):
        if 'from' in self.value:
            module_name, from_ = [v.strip() for v in self.value.split('from')]
            module = importlib.import_module(module_name)
            if module and hasattr(module, from_):
                return getattr(module, from_)
            raise ImportError('Cannot import %s' % self.value)
        else:
            return importlib.import_module(self.value)
